@@ -1,40 +1,43 @@
-# -*- coding: utf-8 -*-
 import functools
-import random
+import logging
 import time
-import warnings
-from collections import namedtuple
+from types import TracebackType
+from typing import Any
+from typing import Callable
+from typing import cast
+from typing import Dict
+from typing import Optional
+from typing import Tuple
+from typing import Type
+from typing import TypeVar
 
 from py_zipkin import Encoding
 from py_zipkin import Kind
 from py_zipkin import storage
 from py_zipkin.encoding._helpers import create_endpoint
-from py_zipkin.encoding._helpers import SpanBuilder
+from py_zipkin.encoding._helpers import Endpoint
+from py_zipkin.encoding._helpers import Span
 from py_zipkin.exception import ZipkinError
+from py_zipkin.logging_helper import TransportHandler
 from py_zipkin.logging_helper import ZipkinLoggingContext
-from py_zipkin.storage import ThreadLocalStack
-from py_zipkin.util import generate_random_128bit_string
+from py_zipkin.request_helpers import create_http_headers
+from py_zipkin.storage import get_default_tracer
+from py_zipkin.storage import SpanStorage
+from py_zipkin.storage import Stack
+from py_zipkin.storage import Tracer
+from py_zipkin.util import create_attrs_for_span
 from py_zipkin.util import generate_random_64bit_string
+from py_zipkin.util import ZipkinAttrs
+
+log = logging.getLogger(__name__)
+
+ERROR_KEY = "error"
 
 
-"""
-Holds the basic attributes needed to log a zipkin trace
-
-:param trace_id: Unique trace id
-:param span_id: Span Id of the current request span
-:param parent_span_id: Parent span Id of the current request span
-:param flags: stores flags header. Currently unused
-:param is_sampled: pre-computed boolean whether the trace should be logged
-"""
-ZipkinAttrs = namedtuple(
-    'ZipkinAttrs',
-    ['trace_id', 'span_id', 'parent_span_id', 'flags', 'is_sampled'],
-)
-
-ERROR_KEY = 'error'
+F = TypeVar("F", bound=Callable[..., Any])
 
 
-class zipkin_span(object):
+class zipkin_span:
     """Context manager/decorator for all of your zipkin tracing needs.
 
     Usage #1: Start a trace with a given sampling rate
@@ -93,27 +96,28 @@ class zipkin_span(object):
 
     def __init__(
         self,
-        service_name,
-        span_name='span',
-        zipkin_attrs=None,
-        transport_handler=None,
-        max_span_batch_size=None,
-        annotations=None,
-        binary_annotations=None,
-        port=0,
-        sample_rate=None,
-        include=None,
-        add_logging_annotation=False,
-        report_root_timestamp=False,
-        use_128bit_trace_id=False,
-        host=None,
-        context_stack=None,
-        span_storage=None,
-        firehose_handler=None,
-        kind=None,
-        timestamp=None,
-        duration=None,
-        encoding=Encoding.V1_THRIFT,
+        service_name: str,
+        span_name: str = "span",
+        zipkin_attrs: Optional[ZipkinAttrs] = None,
+        transport_handler: Optional[TransportHandler] = None,
+        max_span_batch_size: Optional[int] = None,
+        annotations: Optional[Dict[str, Optional[float]]] = None,
+        binary_annotations: Optional[Dict[str, Optional[str]]] = None,
+        port: int = 0,
+        sample_rate: Optional[float] = None,
+        include: Optional[str] = None,
+        add_logging_annotation: bool = False,
+        report_root_timestamp: bool = False,
+        use_128bit_trace_id: bool = False,
+        host: Optional[str] = None,
+        context_stack: Optional[Stack] = None,
+        span_storage: Optional[SpanStorage] = None,
+        firehose_handler: Optional[TransportHandler] = None,
+        kind: Optional[Kind] = None,
+        timestamp: Optional[float] = None,
+        duration: Optional[float] = None,
+        encoding: Encoding = Encoding.V2_JSON,
+        _tracer: Optional[Tracer] = None,
     ):
         """Logs a zipkin span. If this is the root span, then a zipkin
         trace is started as well.
@@ -182,12 +186,15 @@ class zipkin_span(object):
         :param duration: Duration in seconds, defaults to the time spent in the
             context. Set this if you want to use a custom duration.
         :type duration: float
-        :param encoding: Output encoding format, defaults to V1 thrift spans.
+        :param encoding: Output encoding format, defaults to V2_JSON spans.
         :type encoding: Encoding
+        :param _tracer: Current tracer object. This argument is passed in
+            automatically when you create a zipkin_span from a Tracer.
+        :type _tracer: Tracer
         """
         self.service_name = service_name
         self.span_name = span_name
-        self.zipkin_attrs = zipkin_attrs
+        self.zipkin_attrs_override = zipkin_attrs
         self.transport_handler = transport_handler
         self.max_span_batch_size = max_span_batch_size
         self.annotations = annotations or {}
@@ -198,23 +205,22 @@ class zipkin_span(object):
         self.report_root_timestamp_override = report_root_timestamp
         self.use_128bit_trace_id = use_128bit_trace_id
         self.host = host
-        self._context_stack = context_stack or ThreadLocalStack()
-        if span_storage is not None:
-            self._span_storage = span_storage
-        else:
-            self._span_storage = storage.default_span_storage()
+        self._context_stack = context_stack
+        self._span_storage = span_storage
         self.firehose_handler = firehose_handler
         self.kind = self._generate_kind(kind, include)
         self.timestamp = timestamp
         self.duration = duration
         self.encoding = encoding
+        self._tracer = _tracer
 
         self._is_local_root_span = False
-        self.logging_context = None
+        self.logging_context: Optional[ZipkinLoggingContext] = None
         self.do_pop_attrs = False
         # Spans that log a 'cs' timestamp can additionally record a
         # 'sa' binary annotation that shows where the request is going.
-        self.sa_endpoint = None
+        self.remote_endpoint: Optional[Endpoint] = None
+        self.zipkin_attrs: Optional[ZipkinAttrs] = None
 
         # It used to  be possible to override timestamp and duration by passing
         # in the cs/cr or sr/ss annotations. We want to keep backward compatibility
@@ -222,30 +228,31 @@ class zipkin_span(object):
         # same way.
         # This doesn't fit well with v2 spans since those annotations are gone, so
         # we also log a deprecation warning.
-        if 'sr' in self.annotations and 'ss' in self.annotations:
-            self.duration = self.annotations['ss'] - self.annotations['sr']
-            self.timestamp = self.annotations['sr']
-            warnings.warn(
+        if "sr" in self.annotations and "ss" in self.annotations:
+            assert self.annotations["ss"] is not None
+            assert self.annotations["sr"] is not None
+            self.duration = self.annotations["ss"] - self.annotations["sr"]
+            self.timestamp = self.annotations["sr"]
+            log.warning(
                 "Manually setting 'sr'/'ss' annotations is deprecated. Please "
-                "use the timestamp and duration parameters.",
-                DeprecationWarning,
+                "use the timestamp and duration parameters."
             )
-        if 'cr' in self.annotations and 'cs' in self.annotations:
-            self.duration = self.annotations['cr'] - self.annotations['cs']
-            self.timestamp = self.annotations['cs']
-            warnings.warn(
+        if "cr" in self.annotations and "cs" in self.annotations:
+            assert self.annotations["cr"] is not None
+            assert self.annotations["cs"] is not None
+            self.duration = self.annotations["cr"] - self.annotations["cs"]
+            self.timestamp = self.annotations["cs"]
+            log.warning(
                 "Manually setting 'cr'/'cs' annotations is deprecated. Please "
-                "use the timestamp and duration parameters.",
-                DeprecationWarning,
+                "use the timestamp and duration parameters."
             )
 
-        # Root spans have transport_handler and at least one of zipkin_attrs
-        # or sample_rate.
-        if self.zipkin_attrs or self.sample_rate is not None:
+        # Root spans have transport_handler and at least one of
+        # zipkin_attrs_override or sample_rate.
+        if self.zipkin_attrs_override or self.sample_rate is not None:
             # transport_handler is mandatory for root spans
             if self.transport_handler is None:
-                raise ZipkinError(
-                    'Root spans require a transport handler to be given')
+                raise ZipkinError("Root spans require a transport handler to be given")
 
             self._is_local_root_span = True
 
@@ -254,24 +261,40 @@ class zipkin_span(object):
             self._is_local_root_span = True
 
         if self.sample_rate is not None and not (0.0 <= self.sample_rate <= 100.0):
-            raise ZipkinError('Sample rate must be between 0.0 and 100.0')
+            raise ZipkinError("Sample rate must be between 0.0 and 100.0")
 
-        if not isinstance(self._span_storage, storage.SpanStorage):
-            raise ZipkinError('span_storage should be an instance '
-                              'of py_zipkin.storage.SpanStorage')
+        if self._span_storage is not None and not isinstance(
+            self._span_storage, storage.SpanStorage
+        ):
+            raise ZipkinError(
+                "span_storage should be an instance of py_zipkin.storage.SpanStorage"
+            )
 
-    def __call__(self, f):
+        if self._span_storage is not None:
+            log.warning("span_storage is deprecated. Set local_storage instead.")
+            self.get_tracer()._span_storage = self._span_storage
+
+        if self._context_stack is not None:
+            log.warning("context_stack is deprecated. Set local_storage instead.")
+            self.get_tracer()._context_stack = self._context_stack
+
+    def __call__(self, f: F) -> F:
         @functools.wraps(f)
-        def decorated(*args, **kwargs):
+        def decorated(*args: Any, **kwargs: Any) -> Any:
             with zipkin_span(
                 service_name=self.service_name,
                 span_name=self.span_name,
                 zipkin_attrs=self.zipkin_attrs,
                 transport_handler=self.transport_handler,
+                max_span_batch_size=self.max_span_batch_size,
                 annotations=self.annotations,
                 binary_annotations=self.binary_annotations,
                 port=self.port,
                 sample_rate=self.sample_rate,
+                include=None,
+                add_logging_annotation=self.add_logging_annotation,
+                report_root_timestamp=self.report_root_timestamp_override,
+                use_128bit_trace_id=self.use_128bit_trace_id,
                 host=self.host,
                 context_stack=self._context_stack,
                 span_storage=self._span_storage,
@@ -280,14 +303,22 @@ class zipkin_span(object):
                 timestamp=self.timestamp,
                 duration=self.duration,
                 encoding=self.encoding,
+                _tracer=self._tracer,
             ):
                 return f(*args, **kwargs)
-        return decorated
 
-    def __enter__(self):
+        return cast(F, decorated)
+
+    def get_tracer(self) -> Tracer:
+        if self._tracer is not None:
+            return self._tracer
+        else:
+            return get_default_tracer()
+
+    def __enter__(self) -> "zipkin_span":
         return self.start()
 
-    def _generate_kind(self, kind, include):
+    def _generate_kind(self, kind: Optional[Kind], include: Optional[str]) -> Kind:
         # If `kind` is not set, then we generate it from `include`.
         # This code maintains backward compatibility with old versions of py_zipkin
         # which used include rather than kind to identify client / server spans.
@@ -299,13 +330,10 @@ class zipkin_span(object):
                 # than it's a client or server span respectively.
                 # If neither or both are present, then it's a local span
                 # which is represented by kind = None.
-                warnings.warn(
-                    'The include argument is deprecated. Please use kind.',
-                    DeprecationWarning,
-                )
-                if 'client' in include and 'server' not in include:
+                log.warning("The include argument is deprecated. Please use kind.")
+                if "client" in include and "server" not in include:
                     return Kind.CLIENT
-                elif 'client' not in include and 'server' in include:
+                elif "client" not in include and "server" in include:
                     return Kind.SERVER
                 else:
                     return Kind.LOCAL
@@ -313,19 +341,12 @@ class zipkin_span(object):
         # If both kind and include are unset, then it's a local span.
         return Kind.LOCAL
 
-    def start(self):
-        """Enter the new span context. All annotations logged inside this
-        context will be attributed to this span. All new spans generated
-        inside this context will have this span as their parent.
+    def _get_current_context(self) -> Tuple[bool, Optional[ZipkinAttrs]]:
+        """Returns the current ZipkinAttrs and generates new ones if needed.
 
-        In the unsampled case, this context still generates new span IDs and
-        pushes them onto the threadlocal stack, so downstream services calls
-        made will pass the correct headers. However, the logging handler is
-        never attached in the unsampled case, so the spans are never logged.
+        :returns: (report_root_timestamp, zipkin_attrs)
+        :rtype: (bool, ZipkinAttrs)
         """
-        self.do_pop_attrs = False
-        report_root_timestamp = False
-
         # This check is technically not necessary since only root spans will have
         # sample_rate, zipkin_attrs or a transport set. But it helps making the
         # code clearer by separating the logic for a root span from the one for a
@@ -344,51 +365,79 @@ class zipkin_span(object):
             if self.sample_rate is not None:
 
                 # If this trace is not sampled, we re-roll the dice.
-                if self.zipkin_attrs and not self.zipkin_attrs.is_sampled:
+                if (
+                    self.zipkin_attrs_override
+                    and not self.zipkin_attrs_override.is_sampled
+                ):
                     # This will be the root span of the trace, so we should
                     # set timestamp and duration.
-                    report_root_timestamp = True
-                    self.zipkin_attrs = create_attrs_for_span(
-                        sample_rate=self.sample_rate,
-                        trace_id=self.zipkin_attrs.trace_id,
-                        use_128bit_trace_id=self.use_128bit_trace_id,
+                    return (
+                        True,
+                        create_attrs_for_span(
+                            sample_rate=self.sample_rate,
+                            trace_id=self.zipkin_attrs_override.trace_id,
+                        ),
                     )
 
-                # If zipkin_attrs was not passed in, we simply generate new
-                # zipkin_attrs to start a new trace.
-                elif not self.zipkin_attrs:
-                    # This will be the root span of the trace, so we should
-                    # set timestamp and duration.
-                    report_root_timestamp = True
-                    self.zipkin_attrs = create_attrs_for_span(
-                        sample_rate=self.sample_rate,
-                        use_128bit_trace_id=self.use_128bit_trace_id,
+                # If zipkin_attrs_override was not passed in, we simply generate
+                # new zipkin_attrs to start a new trace.
+                elif not self.zipkin_attrs_override:
+                    return (
+                        True,
+                        create_attrs_for_span(
+                            sample_rate=self.sample_rate,
+                            use_128bit_trace_id=self.use_128bit_trace_id,
+                        ),
                     )
 
-            if self.firehose_handler and not self.zipkin_attrs:
+            if self.firehose_handler and not self.zipkin_attrs_override:
                 # If it has gotten here, the only thing that is
                 # causing a trace is the firehose. So we force a trace
                 # with sample rate of 0
-                report_root_timestamp = True
-                self.zipkin_attrs = create_attrs_for_span(
-                    sample_rate=0.0,
-                    use_128bit_trace_id=self.use_128bit_trace_id,
+                return (
+                    True,
+                    create_attrs_for_span(
+                        sample_rate=0.0,
+                        use_128bit_trace_id=self.use_128bit_trace_id,
+                    ),
                 )
+
+            # If we arrive here it means the sample_rate was not set while
+            # zipkin_attrs_override was, so let's simply return that.
+            return False, self.zipkin_attrs_override
+
         else:
-            # If zipkin_attrs was not passed in, we check if there's already a
-            # trace context in _context_stack.
-            if not self.zipkin_attrs:
-                existing_zipkin_attrs = self._context_stack.get()
-                # If there's an existing context, let's create new zipkin_attrs
-                # with that context as parent.
-                if existing_zipkin_attrs:
-                    self.zipkin_attrs = ZipkinAttrs(
+            # Check if there's already a trace context in _context_stack.
+            existing_zipkin_attrs = self.get_tracer().get_zipkin_attrs()
+            # If there's an existing context, let's create new zipkin_attrs
+            # with that context as parent.
+            if existing_zipkin_attrs:
+                return (
+                    False,
+                    ZipkinAttrs(
                         trace_id=existing_zipkin_attrs.trace_id,
                         span_id=generate_random_64bit_string(),
                         parent_span_id=existing_zipkin_attrs.span_id,
                         flags=existing_zipkin_attrs.flags,
                         is_sampled=existing_zipkin_attrs.is_sampled,
-                    )
+                    ),
+                )
+
+        return False, None
+
+    def start(self) -> "zipkin_span":
+        """Enter the new span context. All annotations logged inside this
+        context will be attributed to this span. All new spans generated
+        inside this context will have this span as their parent.
+
+        In the unsampled case, this context still generates new span IDs and
+        pushes them onto the threadlocal stack, so downstream services calls
+        made will pass the correct headers. However, the logging handler is
+        never attached in the unsampled case, so the spans are never logged.
+        """
+        self.do_pop_attrs = False
+
+        report_root_timestamp, self.zipkin_attrs = self._get_current_context()
 
         # If zipkin_attrs are not set up by now, that means this span is not
         # configured to perform logging itself, and it's not in an existing
@@ -397,7 +446,7 @@ class zipkin_span(object):
         if not self.zipkin_attrs:
             return self
 
-        self._context_stack.push(self.zipkin_attrs)
+        self.get_tracer().push_zipkin_attrs(self.zipkin_attrs)
         self.do_pop_attrs = True
 
         self.start_timestamp = time.time()
@@ -406,6 +455,15 @@ class zipkin_span(object):
             # Don't set up any logging if we're not sampling
             if not self.zipkin_attrs.is_sampled and not self.firehose_handler:
                 return self
+            # If transport is already configured don't override it. Doing so would
+            # cause all previously recorded spans to never be emitted as exiting
+            # the inner logging context will reset transport_configured to False.
+            if self.get_tracer().is_transport_configured():
+                log.info(
+                    "Transport was already configured, ignoring override "
+                    "from span {}".format(self.span_name)
+                )
+                return self
             endpoint = create_endpoint(self.port, self.service_name, self.host)
             self.logging_context = ZipkinLoggingContext(
                 self.zipkin_attrs,
@@ -413,7 +471,7 @@ class zipkin_span(object):
                 self.span_name,
                 self.transport_handler,
                 report_root_timestamp or self.report_root_timestamp_override,
-                self._span_storage,
+                self.get_tracer,
                 self.service_name,
                 binary_annotations=self.binary_annotations,
                 add_logging_annotation=self.add_logging_annotation,
@@ -421,16 +479,27 @@ class zipkin_span(object):
                 max_span_batch_size=self.max_span_batch_size,
                 firehose_handler=self.firehose_handler,
                 encoding=self.encoding,
+                annotations=self.annotations,
             )
             self.logging_context.start()
-            self._span_storage.set_transport_configured(configured=True)
+            self.get_tracer().set_transport_configured(configured=True)
 
         return self
 
-    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
+    def __exit__(
+        self,
+        _exc_type: Optional[Type[BaseException]],
+        _exc_value: Optional[BaseException],
+        _exc_traceback: TracebackType,
+    ) -> None:
         self.stop(_exc_type, _exc_value, _exc_traceback)
 
-    def stop(self, _exc_type=None, _exc_value=None, _exc_traceback=None):
+    def stop(
+        self,
+        _exc_type: Optional[Type[BaseException]] = None,
+        _exc_value: Optional[BaseException] = None,
+        _exc_traceback: Optional[TracebackType] = None,
+    ) -> None:
         """Exit the span context. Zipkin attrs are pushed onto the
         threadlocal stack regardless of sampling, so they always need to be
         popped off. The actual logging of spans depends on sampling and that
@@ -438,29 +507,39 @@ class zipkin_span(object):
         """
 
         if self.do_pop_attrs:
-            self._context_stack.pop()
+            self.get_tracer().pop_zipkin_attrs()
 
         # If no transport is configured, there's no reason to create a new Span.
         # This also helps avoiding memory leaks since without a transport nothing
-        # would pull spans out of _span_storage.
-        if not self._span_storage.is_transport_configured():
+        # would pull spans out of get_tracer().
+        if not self.get_tracer().is_transport_configured():
             return
 
         # Add the error annotation if an exception occurred
         if any((_exc_type, _exc_value, _exc_traceback)):
-            error_msg = u'{0}: {1}'.format(_exc_type.__name__, _exc_value)
-            self.update_binary_annotations({
-                ERROR_KEY: error_msg,
-            })
+            assert _exc_type is not None
+            try:
+                error_msg = f"{_exc_type.__name__}: {_exc_value}"
+            except TypeError:
+                # This sometimes happens when an exception raises when calling
+                # __str__ on it.
+                error_msg = f"{_exc_type.__name__}: {_exc_value!r}"
+            self.update_binary_annotations({ERROR_KEY: error_msg})
 
         # Logging context is only initialized for "root" spans of the local
         # process (i.e. this zipkin_span not inside of any other local
         # zipkin_spans)
         if self.logging_context:
-            self.logging_context.stop()
-            self.logging_context = None
-            self._span_storage.set_transport_configured(configured=False)
-            return
+            try:
+                self.logging_context.stop()
+            except Exception as ex:
+                err_msg = f"Error emitting zipkin trace. {repr(ex)}"
+                log.error(err_msg)
+            finally:
+                self.logging_context = None
+                self.get_tracer().clear()
+                self.get_tracer().set_transport_configured(configured=False)
+                return
 
         # If we've gotten here, that means that this span is a child span of
         # this context's root span (i.e. it's a zipkin_span inside another
@@ -472,27 +551,28 @@ class zipkin_span(object):
         else:
             duration = end_timestamp - self.start_timestamp
 
-        self._span_storage.append(SpanBuilder(
-            trace_id=self.zipkin_attrs.trace_id,
-            name=self.span_name,
-            parent_id=self.zipkin_attrs.parent_span_id,
-            span_id=self.zipkin_attrs.span_id,
-            timestamp=self.timestamp if self.timestamp else self.start_timestamp,
-            duration=duration,
-            annotations=self.annotations,
-            tags=self.binary_annotations,
-            kind=self.kind,
-            service_name=self.service_name,
-            sa_endpoint=self.sa_endpoint,
-        ))
+        endpoint = create_endpoint(self.port, self.service_name, self.host)
+        assert self.zipkin_attrs is not None
+        self.get_tracer().add_span(
+            Span(
+                trace_id=self.zipkin_attrs.trace_id,
+                name=self.span_name,
+                parent_id=self.zipkin_attrs.parent_span_id,
+                span_id=self.zipkin_attrs.span_id,
+                kind=self.kind,
+                timestamp=self.timestamp if self.timestamp else self.start_timestamp,
+                duration=duration,
+                annotations=self.annotations,
+                local_endpoint=endpoint,
+                remote_endpoint=self.remote_endpoint,
+                tags=self.binary_annotations,
+            )
+        )
 
-    def update_binary_annotations(self, extra_annotations):
-        """Updates the binary annotations for the current span.
-
-        If this trace is not being sampled then this is a no-op.
-        """
-        if not self.zipkin_attrs:
-            return
+    def update_binary_annotations(
+        self, extra_annotations: Dict[str, Optional[str]]
+    ) -> None:
+        """Updates the binary annotations for the current span."""
         if not self.logging_context:
             # This is not the root span, so binary annotations will be added
             # to the log handler when this span context exits.
@@ -500,14 +580,34 @@ class zipkin_span(object):
         else:
             # Otherwise, we're in the context of the root span, so just update
             # the binary annotations for the logging context directly.
-            self.logging_context.binary_annotations_dict.update(extra_annotations)
+            self.logging_context.tags.update(extra_annotations)
+
+    def add_annotation(self, value: str, timestamp: Optional[float] = None) -> None:
+        """Add an annotation for the current span
+
+        The timestamp defaults to "now", but may be specified.
+
+        :param value: The annotation string
+        :type value: str
+        :param timestamp: Timestamp for the annotation
+        :type timestamp: float
+        """
+        timestamp = timestamp or time.time()
+        if not self.logging_context:
+            # This is not the root span, so annotations will be added
+            # to the log handler when this span context exits.
+            self.annotations[value] = timestamp
+        else:
+            # Otherwise, we're in the context of the root span, so just update
+            # the annotations for the logging context directly.
+            self.logging_context.annotations[value] = timestamp
 
     def add_sa_binary_annotation(
         self,
-        port=0,
-        service_name='unknown',
-        host='127.0.0.1',
-    ):
+        port: int = 0,
+        service_name: str = "unknown",
+        host: str = "127.0.0.1",
+    ) -> None:
         """Adds a 'sa' binary annotation to the current span.
 
         'sa' binary annotations are useful for situations where you need to log
@@ -522,29 +622,26 @@ class zipkin_span(object):
         :param host: Host address of the destination
         :type host: str
         """
-        if not self.zipkin_attrs:
-            return
-
         if self.kind != Kind.CLIENT:
             # TODO: trying to set a sa binary annotation for a non-client span
             # should result in a logged error
             return
 
-        sa_endpoint = create_endpoint(
+        remote_endpoint = create_endpoint(
             port=port,
             service_name=service_name,
             host=host,
         )
         if not self.logging_context:
-            if self.sa_endpoint is not None:
-                raise ValueError('SA annotation already set.')
-            self.sa_endpoint = sa_endpoint
+            if self.remote_endpoint is not None:
+                raise ValueError("SA annotation already set.")
+            self.remote_endpoint = remote_endpoint
         else:
-            if self.logging_context.sa_endpoint is not None:
-                raise ValueError('SA annotation already set.')
-            self.logging_context.sa_endpoint = sa_endpoint
+            if self.logging_context.remote_endpoint is not None:
+                raise ValueError("SA annotation already set.")
+            self.logging_context.remote_endpoint = remote_endpoint
 
-    def override_span_name(self, name):
+    def override_span_name(self, name: str) -> None:
         """Overrides the current span name.
 
         This is useful if you don't know the span name yet when you create the
@@ -560,11 +657,11 @@ class zipkin_span(object):
             self.logging_context.span_name = name
 
 
-def _validate_args(kwargs):
-    if 'kind' in kwargs:
+def _validate_args(kwargs: Dict[str, Any]) -> None:
+    if "kind" in kwargs:
         raise ValueError(
             '"kind" is not valid in this context. '
-            'You probably want to use zipkin_span()'
+            "You probably want to use zipkin_span()"
         )
 
 
@@ -574,15 +671,15 @@ class zipkin_client_span(zipkin_span):
     Subclass of :class:`zipkin_span` using only annotations relevant to clients
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         """Logs a zipkin span with client annotations.
 
         See :class:`zipkin_span` for arguments
         """
         _validate_args(kwargs)
 
-        kwargs['kind'] = Kind.CLIENT
-        super(zipkin_client_span, self).__init__(*args, **kwargs)
+        kwargs["kind"] = Kind.CLIENT
+        super().__init__(*args, **kwargs)
 
 
 class zipkin_server_span(zipkin_span):
@@ -591,59 +688,20 @@ class zipkin_server_span(zipkin_span):
     Subclass of :class:`zipkin_span` using only annotations relevant to servers
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         """Logs a zipkin span with server annotations.
 
         See :class:`zipkin_span` for arguments
         """
         _validate_args(kwargs)
 
-        kwargs['kind'] = Kind.SERVER
-        super(zipkin_server_span, self).__init__(*args, **kwargs)
+        kwargs["kind"] = Kind.SERVER
+        super().__init__(*args, **kwargs)
 
 
-def create_attrs_for_span(
-    sample_rate=100.0,
-    trace_id=None,
-    span_id=None,
-    use_128bit_trace_id=False,
-):
-    """Creates a set of zipkin attributes for a span.
-
-    :param sample_rate: Float between 0.0 and 100.0 to determine sampling rate
-    :type sample_rate: float
-    :param trace_id: Optional 16-character hex string representing a trace_id.
-                    If this is None, a random trace_id will be generated.
-    :type trace_id: str
-    :param span_id: Optional 16-character hex string representing a span_id.
-                    If this is None, a random span_id will be generated.
-    :type span_id: str
-    :param use_128bit_trace_id: If true, generate 128-bit trace_ids
-    :type use_128bit_trace_id: boolean
-    """
-    # Calculate if this trace is sampled based on the sample rate
-    if trace_id is None:
-        if use_128bit_trace_id:
-            trace_id = generate_random_128bit_string()
-        else:
-            trace_id = generate_random_64bit_string()
-    if span_id is None:
-        span_id = generate_random_64bit_string()
-    if sample_rate == 0.0:
-        is_sampled = False
-    else:
-        is_sampled = (random.random() * 100) < sample_rate
-
-    return ZipkinAttrs(
-        trace_id=trace_id,
-        span_id=span_id,
-        parent_span_id=None,
-        flags='0',
-        is_sampled=is_sampled,
-    )
-
-
-def create_http_headers_for_new_span(context_stack=None):
+def create_http_headers_for_new_span(
+    context_stack: Optional[Stack] = None, tracer: Optional[Tracer] = None
+) -> Dict[str, Optional[str]]:
     """
     Generate the headers for a new zipkin span.
 
@@ -655,16 +713,4 @@ def create_http_headers_for_new_span(context_stack=None):
     :returns: dict containing (X-B3-TraceId, X-B3-SpanId, X-B3-ParentSpanId,
                 X-B3-Flags and X-B3-Sampled) keys OR an empty dict.
     """
-    if context_stack is None:
-        context_stack = ThreadLocalStack()
-    zipkin_attrs = context_stack.get()
-    if not zipkin_attrs:
-        return {}
-
-    return {
-        'X-B3-TraceId': zipkin_attrs.trace_id,
-        'X-B3-SpanId': generate_random_64bit_string(),
-        'X-B3-ParentSpanId': zipkin_attrs.span_id,
-        'X-B3-Flags': '0',
-        'X-B3-Sampled': '1' if zipkin_attrs.is_sampled else '0',
-    }
+    return create_http_headers(context_stack, tracer, True)
